/*
  VFB-01 : Virtual FB-01 emulator

  Copyright 2000 by Daisuke Nagano <breeze.nagano@nifty.ne.jp>
  Feb.10.1999

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _VFB01_H_
#define _VFB01_H_

/* ------------------------------------------------------------------- */

#define VFB_MAX_CHANNEL_NUMBER        8 /* up to 16 ( <= heavy load ) */
#define VFB_INITIAL_VOICE_NUMBER      0

#ifndef MIDI_DEVICE_NAME
# define MIDI_DEVICE_NAME "/dev/midi"
#endif

#ifndef DSP_DEVICE_NAME
# define DSP_DEVICE_NAME "/dev/dsp"
#endif

#ifndef VOICE_PARAMETER_NAME
# define VOICE_PARAMETER_NAME "/usr/local/share/vfb01/voices.cfg"
#endif

/* ------------------------------------------------------------------- */

#define VFB_MAX_FM_SLOTS              8
#define VFB_MAX_TONE_NUMBER         256

#define FLAG_TRUE                     1
#define FLAG_FALSE                    0

#define VFB_PAN_N                     0
#define VFB_PAN_L                     1
#define VFB_PAN_R                     2
#define VFB_PAN_C                     3

#define SYSEX_BUF_SIZE             1024
#define VFB_VERSION_TEXT_SIZE       256

/* ------------------------------------------------------------------- */

typedef struct _MidiEvent {
  int type;
  int ch;
  int a,b;
  int ex_buf[SYSEX_BUF_SIZE];
} MidiEvent;

typedef struct _VOICE_DATA {
  int voice_number;
  int fl;              /* feed back */
  int con;             /* connection:algorithm */
  int slot_mask;       /* slot mask */
  int dt1[4];          /* detune 1 */
  int dt2[4];          /* detune 2 */
  int mul[4];          /* multiple */
  int tl[4];           /* total level */
  int ks[4];           /* key scale */
  int ar[4];           /* attack rate */
  int ame[4];          /* amplitude moduration enable */
  int d1r[4];          /* decay rate 1 */
  int d2r[4];          /* decay rate 2 ( sustine rate ) */
  int rr[4];           /* release rate */
  int sl[4];           /* sustine level */

  int v0;
  int v1[4];
  int v2[4];
  int v3[4];
  int v4[4];
  int v5[4];
  int v6[4];

} VOICE_DATA;

typedef struct _VFB_DATA {
  
  unsigned char version_1[VFB_VERSION_TEXT_SIZE];
  unsigned char version_2[VFB_VERSION_TEXT_SIZE];

  VOICE_DATA voice[VFB_MAX_TONE_NUMBER];

  char *voice_parameter_file;

  /* playing work area */

  long total_count;       /* total steps */
  long elapsed_time;      /* unit = microsecound */

  /* user configuration */

  int master_volume;
  int units;

  /* work parameter */

  int is_normal_exit;

  int is_output_to_stdout;
  int is_use_fragment;
  int is_ignore_ch10;
  int verbose;

  char *dsp_device;
  char *midi_device;

  int  dsp_speed;

} VFB_DATA;

/* ------------------------------------------------------------------- */

extern int openVirtualMidiDevice( const char * );
extern int closeVirtualMidiDevice( void );

extern int vfb01( VFB_DATA * );

extern int getMidiEvent( MidiEvent * );

/* ------------------------------------------------------------------- */

extern VOICE_DATA **voices;

/* ------------------------------------------------------------------- */

#endif /* _VFB01_H_ */
